import classNames from 'classnames'
import { useClient } from 'cozy-client'
import { useI18n } from 'cozy-ui/transpiled/react/I18n'
import Icon from 'cozy-ui/transpiled/react/Icon'
import {
  DataloadSectionType,
  DataloadState,
  FluidType,
  UsageEventType,
} from 'enums'
import { Dataload } from 'models'
import React, { useCallback } from 'react'
import { NavLink } from 'react-router-dom'
import ConverterService from 'services/converter.service'
import UsageEventService from 'services/usageEvent.service'
import { getNavPicto } from 'utils/picto'
import { formatNumberValues, getFluidName } from 'utils/utils'

interface DataloadSectionDetailProps {
  dataload: Dataload
  fluidType: FluidType
  dataloadSectionType: DataloadSectionType
}

const DataloadSectionDetail = ({
  dataload,
  fluidType,
  dataloadSectionType,
}: DataloadSectionDetailProps) => {
  const client = useClient()
  const { t } = useI18n()
  const converterService = new ConverterService()

  const isMulti = fluidType === FluidType.MULTIFLUID
  const isCompare = dataloadSectionType === DataloadSectionType.NO_COMPARE

  const emitNavEvent = useCallback(
    async (targetPage: string) => {
      await UsageEventService.addEvent(client, {
        type: UsageEventType.NAVIGATION_EVENT,
        target: targetPage,
      })
    },
    [client]
  )

  if (!isMulti) {
    return (
      <div
        className={classNames('dataloadvisualizer-euro text-16-normal', {
          [`${getFluidName(fluidType)}`]:
            dataloadSectionType !== DataloadSectionType.LEFT,
          [`${getFluidName(fluidType)}-compare`]:
            dataloadSectionType === DataloadSectionType.LEFT,
        })}
      >
        {`${formatNumberValues(
          converterService.LoadToEuro(
            dataload.value,
            fluidType,
            dataload.price ? dataload.price : null
          )
        )} €`}
      </div>
    )
  }

  if (isMulti && !isCompare) {
    return <div className="dataloadvisualizer-euro text-16-normal" />
  }

  if (isMulti && isCompare && dataload.valueDetail) {
    return (
      <div className="dataloadvisualizer-euro text-16-normal">
        {dataload.valueDetail.map((value, index) => {
          const isValid = value.state === DataloadState.VALID
          const isUpcoming = value.state === DataloadState.UPCOMING
          const isComing = value.state === DataloadState.COMING
          const isEmpty = value.state === DataloadState.EMPTY
          const isHole = value.state === DataloadState.HOLE
          const isMissing = value.state === DataloadState.MISSING
          return (
            <NavLink
              key={FluidType[index]}
              to={`/consumption/${FluidType[index].toLowerCase()}`}
              className="dataloadvisualizer-euro-link"
              onClick={() => emitNavEvent(FluidType[index].toLowerCase())}
            >
              <div
                className={classNames('dataloadvisualizer-euro-fluid', {
                  [FluidType[index].toLowerCase()]:
                    isValid || isUpcoming || isComing || isEmpty || isHole,
                  error: isMissing,
                })}
              >
                <Icon
                  className="dataloadvisualizer-euro-fluid-icon"
                  icon={getNavPicto(index, true, true)}
                  size={22}
                />
                <div>
                  {isValid && `${formatNumberValues(value.value)} €`}
                  {(isUpcoming || isComing) &&
                    t('consumption_visualizer.data_to_come')}
                  {(isEmpty || isHole) &&
                    t('consumption_visualizer.data_empty')}
                  {isMissing && t('consumption_visualizer.aie')}
                </div>
              </div>
            </NavLink>
          )
        })}
      </div>
    )
  }
  return (
    <div className="dataloadvisualizer-euro text-16-normal">
      <NavLink
        to={`/consumption/${getFluidName(fluidType)}`}
        className="dataloadvisualizer-euro-link"
      >
        <div
          className={`dataloadvisualizer-euro-fluid ${getFluidName(fluidType)}`}
        >
          <Icon
            className="dataloadvisualizer-euro-fluid-icon"
            icon={getNavPicto(fluidType, true, true)}
            size={22}
          />
          <div>{`${formatNumberValues(
            converterService.LoadToEuro(
              dataload.value,
              fluidType,
              dataload.price ? dataload.price : null
            )
          )} €`}</div>
        </div>
      </NavLink>
    </div>
  )
}

export default DataloadSectionDetail
