/* eslint-disable camelcase */
import { QueryResult } from 'cozy-client'
import {
  DataloadState,
  UserActionState,
  UserChallengeState,
  UserChallengeSuccess,
  UserChallengeUpdateFlag,
  UserDuelState,
} from 'enums'
import { cloneDeep } from 'lodash'
import { DateTime, Duration } from 'luxon'
import {
  ChallengeEntity,
  Dataload,
  DuelEntity,
  ExplorationEntity,
  QuizEntity,
  UserChallenge,
} from 'models'
import {
  allChallengeEntityData,
  challengeEntityData,
} from 'tests/__mocks__/challengeEntity.mock'
import { graphData } from 'tests/__mocks__/chartData.mock'
import mockClient from 'tests/__mocks__/client.mock'
import {
  allDuelEntity,
  duelData,
  duelEntity,
} from 'tests/__mocks__/duelData.mock'
import {
  explorationDefault,
  explorationEntity,
  explorationEntity4,
  explorationEntity4_0,
  userExploration,
  userExploration1,
  userExploration4,
  userExploration4_0,
  userExplorationUnlocked,
} from 'tests/__mocks__/explorationData.mock'
import { fluidStatusData } from 'tests/__mocks__/fluidStatusData.mock'
import { quizEntity, userQuiz } from 'tests/__mocks__/quizData.mock'
import { getError } from 'tests/__mocks__/testUtils'
import {
  userChallengeData,
  userChallengeDefault,
  userChallengeExplo1OnGoing,
  userChallengeExplo4,
  userChallengeExplo4_0,
} from 'tests/__mocks__/userChallengeData.mock'
import ChallengeService from './challenge.service'

const mockGetExplorationEntityById = jest.fn()
const mockParseExplorationEntityToUserExploration = jest.fn()
const mockGetUserExplorationFromExplorationEntities = jest.fn()
jest.mock('./exploration.service', () => {
  return jest.fn(() => ({
    getExplorationEntityById: mockGetExplorationEntityById,
    parseExplorationEntityToUserExploration:
      mockParseExplorationEntityToUserExploration,
    getUserExplorationfromExplorationEntities:
      mockGetUserExplorationFromExplorationEntities,
  }))
})

const mockGetGraphData = jest.fn()
const mockCalculatePerformanceIndicatorValue = jest.fn()
jest.mock('./consumption.service', () => {
  return jest.fn(() => ({
    getGraphData: mockGetGraphData,
    calculatePerformanceIndicatorValue: mockCalculatePerformanceIndicatorValue,
  }))
})

describe('Challenge service', () => {
  const challengeService = new ChallengeService(mockClient)
  beforeEach(() => {
    jest.clearAllMocks()
  })

  describe('unLockCurrentUserChallenge method', () => {
    it('should return all user challenge', () => {
      const result =
        challengeService.unLockCurrentUserChallenge(userChallengeData)
      expect(result).toEqual(userChallengeData)
    })
    it('should return the challenge unlocked', () => {
      const expectedResult = userChallengeData[1]
      expectedResult.state = UserChallengeState.UNLOCKED
      const result = challengeService.unLockCurrentUserChallenge([
        userChallengeData[1],
      ])
      expect(result).toEqual([expectedResult])
    })
  })

  describe('parseChallengeEntityToUserChallenge method', () => {
    it('should return a user challenge', () => {
      const expectedResult: UserChallenge = {
        id: challengeEntityData.id,
        title: challengeEntityData.title,
        description: challengeEntityData.description,
        target: challengeEntityData.target,
        duel: duelData,
        state: UserChallengeState.LOCKED,
        progress: {
          quizProgress: 0,
          explorationProgress: 0,
          actionProgress: 0,
        },
        success: UserChallengeSuccess.ONGOING,
        startDate: null,
        endingDate: null,
        quiz: userQuiz,
        exploration: userExplorationUnlocked,
        action: {
          ecogesture: null,
          startDate: null,
          state: UserActionState.UNSTARTED,
        },
      }
      const result = challengeService.parseChallengeEntityToUserChallenge(
        challengeEntityData,
        duelData,
        userQuiz,
        userExplorationUnlocked
      )
      expect(result).toEqual(expectedResult)
    })
  })

  describe('buildUserChallengeList method', () => {
    const mockQueryResult: QueryResult<ChallengeEntity[], DuelEntity[]> = {
      data: allChallengeEntityData,
      included: [duelEntity],
      bookmark: '',
      next: false,
      skip: 0,
    }
    const mockQueryResultQuiz: QueryResult<ChallengeEntity[], QuizEntity[]> = {
      data: allChallengeEntityData,
      included: [quizEntity],
      bookmark: '',
      next: false,
      skip: 0,
    }
    const mockQueryResultExploration: QueryResult<
      ChallengeEntity[],
      ExplorationEntity[]
    > = {
      data: allChallengeEntityData,
      included: [explorationEntity],
      bookmark: '',
      next: false,
      skip: 0,
    }
    const mockQueryResultUser: QueryResult<UserChallenge[]> = {
      data: userChallengeData,
      bookmark: '',
      next: false,
      skip: 0,
    }
    it('should return all user challenge', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResult)
      mockClient.query.mockResolvedValueOnce(mockQueryResultExploration)
      mockClient.query.mockResolvedValueOnce(mockQueryResultQuiz)
      mockClient.query.mockResolvedValueOnce(mockQueryResultUser)
      mockGetExplorationEntityById.mockResolvedValue(explorationEntity)
      mockParseExplorationEntityToUserExploration.mockReturnValue(
        userExploration
      )
      mockGetUserExplorationFromExplorationEntities.mockReturnValue(
        userExploration
      )
      const result =
        await challengeService.buildUserChallengeList(fluidStatusData)
      expect(result).toEqual(userChallengeData)
    })
    it('should return all user challenge plus create one missing', async () => {
      const _mockQueryResult: QueryResult<ChallengeEntity[], DuelEntity[]> = {
        data: allChallengeEntityData,
        included: allDuelEntity,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(_mockQueryResult)
      mockClient.query.mockResolvedValueOnce(mockQueryResultQuiz)
      mockClient.query.mockResolvedValueOnce(mockQueryResultExploration)
      mockClient.query.mockResolvedValueOnce(mockQueryResultUser)
      const result =
        await challengeService.buildUserChallengeList(fluidStatusData)
      expect(result).toEqual(userChallengeData)
    })
    it('should create all user challenge', async () => {
      const mockQueryResultUserEmpty: QueryResult<UserChallenge[]> = {
        data: [],
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(mockQueryResult)
      mockClient.query.mockResolvedValueOnce(mockQueryResultQuiz)
      mockClient.query.mockResolvedValueOnce(mockQueryResultExploration)
      mockClient.query.mockResolvedValueOnce(mockQueryResultUserEmpty)
      mockGetExplorationEntityById.mockResolvedValue(explorationEntity)
      mockParseExplorationEntityToUserExploration.mockReturnValue(
        explorationDefault
      )
      mockGetUserExplorationFromExplorationEntities.mockReturnValue(
        explorationDefault
      )
      const result =
        await challengeService.buildUserChallengeList(fluidStatusData)
      expect(result).toEqual(userChallengeDefault)
    })
  })

  describe('getAllChallengeEntities method', () => {
    it('should return all Challenges Entities', async () => {
      const mockQueryResult: QueryResult<ChallengeEntity[]> = {
        data: allChallengeEntityData,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(mockQueryResult)
      const result = await challengeService.getAllChallengeEntities()
      expect(result).toEqual(allChallengeEntityData)
    })
  })

  describe('deleteAllChallengeEntities method', () => {
    it('should return delete all Challenges Entities', async () => {
      const mockQueryResult: QueryResult<ChallengeEntity[]> = {
        data: allChallengeEntityData,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(mockQueryResult)
      const result = await challengeService.deleteAllChallengeEntities()
      expect(result).toBeTruthy()
    })
    it('should throw an error because destroy failed', async () => {
      const mockQueryResult: QueryResult<ChallengeEntity[]> = {
        data: allChallengeEntityData,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(mockQueryResult)
      mockClient.destroy.mockRejectedValue(new Error())
      const error = await getError(async () =>
        challengeService.deleteAllChallengeEntities()
      )
      expect(error).toEqual(new Error())
    })
  })

  describe('getAllUserChallengeEntities method', () => {
    it('should return all UserChallenges Entities', async () => {
      const mockQueryResult: QueryResult<UserChallenge[]> = {
        data: userChallengeData,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(mockQueryResult)
      const result = await challengeService.getAllUserChallengeEntities()
      expect(result).toEqual(userChallengeData)
    })
  })

  describe('initChallengeDuelProgress method', () => {
    it('should return updatedUserChallenge and dataload', async () => {
      jest
        .spyOn(challengeService, 'getUserChallengeDataload')
        .mockResolvedValueOnce(graphData.actualData)
      mockCalculatePerformanceIndicatorValue.mockResolvedValueOnce(130.83585)
      const expectedUpdatedUserChallenge: UserChallenge = {
        ...userChallengeData[0],
        duel: {
          ...userChallengeData[0].duel,
          userConsumption: 130.83585,
        },
      }
      const mockQueryResult: QueryResult<UserChallenge> = {
        data: expectedUpdatedUserChallenge,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.save.mockResolvedValue(mockQueryResult)
      const expectedResult = {
        updatedUserChallenge: expectedUpdatedUserChallenge,
        dataloads: graphData.actualData,
      }
      const result = await challengeService.initChallengeDuelProgress(
        userChallengeData[0]
      )
      expect(result).toEqual(expectedResult)
    })
    it('should throw an error because create failed', async () => {
      mockClient.save.mockRejectedValue(new Error())
      const error = await getError(async () =>
        challengeService.initChallengeDuelProgress(userChallengeData[0])
      )
      expect(error).toEqual(new Error())
    })
  })

  describe('startUserChallenge method', () => {
    it('should return all UserChallenges Entities', async () => {
      const mockQueryResult: QueryResult<UserChallenge> = {
        data: userChallengeData[2],
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.create.mockResolvedValueOnce(mockQueryResult)
      const result = await challengeService.startUserChallenge(
        userChallengeData[0]
      )
      expect(result).toEqual(userChallengeData[2])
    })
    it('should throw an error because create failed', async () => {
      mockClient.create.mockRejectedValue(new Error())
      const error = await getError(async () =>
        challengeService.startUserChallenge(userChallengeData[0])
      )
      expect(error).toEqual(new Error())
    })
  })

  describe('updateUserChallenge method', () => {
    it('should return the saved userChallenge', async () => {
      const mockQueryResult: QueryResult<UserChallenge> = {
        data: userChallengeData[0],
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.save.mockResolvedValue(mockQueryResult)
      const result = await challengeService.updateUserChallenge(
        userChallengeData[0],
        UserChallengeUpdateFlag.CHALLENGE
      )
      expect(result).toEqual(userChallengeData[0])
      const resultDefault = await challengeService.updateUserChallenge(
        userChallengeData[0],
        999 as UserChallengeUpdateFlag
      )
      expect(resultDefault).toEqual(userChallengeData[0])
    })
    it('should return the updated with duel start userChallenge', async () => {
      const mockQueryResult: QueryResult<UserChallenge> = {
        data: userChallengeData[0],
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.save.mockResolvedValueOnce(mockQueryResult)
      const result = await challengeService.updateUserChallenge(
        userChallengeData[0],
        UserChallengeUpdateFlag.DUEL_START
      )
      expect(result).toEqual(userChallengeData[0])
    })
    it('should throw error when failed to save the updateUserChallenge', async () => {
      mockClient.save.mockRejectedValueOnce(new Error())
      const error = await getError(async () =>
        challengeService.updateUserChallenge(
          userChallengeData[0],
          UserChallengeUpdateFlag.DUEL_START
        )
      )
      expect(error).toEqual(new Error())
    })
  })

  describe('getUserChallengeDataload method', () => {
    it('should return a dataload list', async () => {
      const updatedUserChallenge = {
        ...userChallengeData[0],
        duel: {
          ...userChallengeData[0].duel,
          startDate: DateTime.local().setZone('utc', {
            keepLocalTime: true,
          }),
        },
      }
      mockGetGraphData.mockResolvedValueOnce(graphData)
      const result =
        await challengeService.getUserChallengeDataload(updatedUserChallenge)
      expect(result).toEqual(graphData.actualData)
    })
    it('should return a empty dataload list when duel has no start date', async () => {
      const result = await challengeService.getUserChallengeDataload(
        userChallengeData[0]
      )
      expect(result).toEqual([])
    })
    it('should return a empty dataload list', async () => {
      mockGetGraphData.mockResolvedValueOnce(null)
      const result = await challengeService.getUserChallengeDataload(
        userChallengeData[0]
      )
      expect(result).toEqual([])
    })
  })

  describe('isChallengeDone method', () => {
    describe('case date + lags + 1 is reached', () => {
      const userChallenge = {
        ...userChallengeData[0],
        state: UserChallengeState.DUEL,
        duel: {
          ...userChallengeData[0].duel,
          state: UserDuelState.ONGOING,
          duration: Duration.fromObject({ day: 3 }),
          threshold: 200,
          userConsumption: 199,
          startDate: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 5 }),
        },
      }
      const dataloads: Dataload[] = [
        {
          date: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 5 }),
          value: 69.18029999999999,
          state: DataloadState.AGGREGATED_VALID,
          valueDetail: [
            { value: 45.127739999999996, state: DataloadState.VALID },
            { value: 0.9048899999999999, state: DataloadState.VALID },
            { value: 23.147669999999998, state: DataloadState.VALID },
          ],
        },
        {
          date: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 4 }),
          value: 61.65554999999999,
          state: DataloadState.AGGREGATED_VALID,
          valueDetail: [
            { value: 40.21918999999999, state: DataloadState.VALID },
            { value: 0.8064649999999999, state: DataloadState.VALID },
            { value: 20.629894999999998, state: DataloadState.VALID },
          ],
        },
        {
          date: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 3 }),
          value: 50.0,
          state: DataloadState.AGGREGATED_VALID,
          valueDetail: [
            { value: 25.0, state: DataloadState.VALID },
            { value: 5.0, state: DataloadState.VALID },
            { value: 20.0, state: DataloadState.VALID },
          ],
        },
      ]
      it('should return isDone = true, isWin = true and isEmpty=false when userConsumption < threshold', async () => {
        const result = await challengeService.isChallengeDone(
          userChallenge,
          dataloads
        )
        expect(result).toEqual({ isDone: true, isWin: true })
      })
      it('should return isDone = true, isWin = false when userConsumption >= threshold', async () => {
        const updatedUserChallenge = {
          ...userChallenge,
          duel: {
            ...userChallenge.duel,
            threshold: 100,
            userConsumption: 200,
          },
        }
        const result = await challengeService.isChallengeDone(
          updatedUserChallenge,
          dataloads
        )
        expect(result).toEqual({ isDone: true, isWin: false })
      })
    })

    describe('case date + lags + 1 is not reached', () => {
      const userChallenge = {
        ...userChallengeData[0],
        state: UserChallengeState.DUEL,
        duel: {
          ...userChallengeData[0].duel,
          state: UserDuelState.ONGOING,
          duration: Duration.fromObject({ day: 3 }),
          threshold: 200,
          userConsumption: 199,
          startDate: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 3 }),
        },
      }
      const dataloads: Dataload[] = [
        {
          date: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 3 }),
          value: 69.18029999999999,
          state: DataloadState.AGGREGATED_VALID,
          valueDetail: [
            { value: 45.127739999999996, state: DataloadState.VALID },
            { value: 0.9048899999999999, state: DataloadState.VALID },
            { value: 23.147669999999998, state: DataloadState.VALID },
          ],
        },
        {
          date: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 2 }),
          value: 61.65554999999999,
          state: DataloadState.AGGREGATED_VALID,
          valueDetail: [
            { value: 40.21918999999999, state: DataloadState.VALID },
            { value: 0.8064649999999999, state: DataloadState.VALID },
            { value: 20.629894999999998, state: DataloadState.VALID },
          ],
        },
        {
          date: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .minus({ days: 1 }),
          value: 50,
          state: DataloadState.AGGREGATED_VALID,
          valueDetail: [
            { value: 25.0, state: DataloadState.VALID },
            { value: 5.0, state: DataloadState.VALID },
            { value: 20.0, state: DataloadState.VALID },
          ],
        },
      ]
      it('should return isDone = true, isWin = true when all data are available and userConsumption < threshold', async () => {
        const result = await challengeService.isChallengeDone(
          userChallenge,
          dataloads
        )
        expect(result).toEqual({ isDone: true, isWin: true })
      })

      it('should return isDone = true and isWin = false when all data are available and userConsumption >= threshold', async () => {
        const updatedUserChallenge = {
          ...userChallenge,
          duel: {
            ...userChallenge.duel,
            threshold: 100,
            userConsumption: 200,
          },
        }
        const result = await challengeService.isChallengeDone(
          updatedUserChallenge,
          dataloads
        )
        expect(result).toEqual({ isDone: true, isWin: false })
      })
      it('should return isDone = false and isWin = false when last data is not available', async () => {
        const updatedDataloads = cloneDeep(dataloads)
        updatedDataloads[2].value = -1
        updatedDataloads[2].valueDetail = null
        const result = await challengeService.isChallengeDone(
          userChallenge,
          updatedDataloads
        )
        expect(result).toEqual({ isDone: false, isWin: false })
      })
      it('should return isDone = false and isWin = false when data in the middle is not available', async () => {
        const updatedDataloads = cloneDeep(dataloads)
        updatedDataloads[1].valueDetail = [
          { value: 20.0, state: DataloadState.VALID },
          { value: -1, state: DataloadState.MISSING },
          { value: 10.0, state: DataloadState.VALID },
        ]
        const result = await challengeService.isChallengeDone(
          userChallenge,
          updatedDataloads
        )
        expect(result).toEqual({ isDone: false, isWin: false })
      })
    })
  })

  describe('loopVerificationExplorationCondition method', () => {
    it('should return updated userChallenge with non-conditional exploration', async () => {
      mockGetExplorationEntityById.mockResolvedValue(explorationEntity)
      mockParseExplorationEntityToUserExploration.mockReturnValue(
        userExploration1
      )
      const result =
        await challengeService.loopVerificationExplorationCondition(
          userChallengeExplo1OnGoing,
          allChallengeEntityData,
          fluidStatusData
        )
      expect(result).toEqual(userChallengeExplo1OnGoing)
    })
    it('should return updated userChallenge with condition-validated exploration', async () => {
      mockGetExplorationEntityById.mockResolvedValue(explorationEntity4)
      mockParseExplorationEntityToUserExploration.mockReturnValue(
        userExploration4
      )

      const result =
        await challengeService.loopVerificationExplorationCondition(
          userChallengeExplo4,
          allChallengeEntityData,
          fluidStatusData
        )
      expect(result).toEqual(userChallengeExplo4)
    })
    it('should return updated userChallenge with new exploration when condition is invalid', async () => {
      mockGetExplorationEntityById.mockResolvedValue(explorationEntity4_0)
      mockParseExplorationEntityToUserExploration.mockReturnValue(
        userExploration4_0
      )
      const result =
        await challengeService.loopVerificationExplorationCondition(
          userChallengeExplo4,
          allChallengeEntityData,
          fluidStatusData
        )
      expect(result).toEqual(userChallengeExplo4_0)
    })
  })
})
