import { QueryResult } from 'cozy-client'
import ecogestureData from 'db/ecogestureData.json'
import { EquipmentType, IndividualOrCollective, WarmingType } from 'enums'
import { Ecogesture } from 'models'
import { ProfileEcogesture } from 'models/profileEcogesture.model'
import mockClient from 'tests/__mocks__/client.mock'
import {
  BoilerEcogesture,
  BoilerEcogestureFalse,
  ecogesturesECSData,
  ecogesturesHeatingData,
  mockedEcogesturesData,
} from 'tests/__mocks__/ecogesturesData.mock'
import { mockProfileEcogesture } from 'tests/__mocks__/profileEcogesture.mock'
import { getError } from 'tests/__mocks__/testUtils'
import { hashFile } from 'utils/hash'
import EcogestureService from './ecogesture.service'

// Add types definition to json data
const ecoData = ecogestureData as Ecogesture[]

const mockQueryResultCreated: QueryResult<boolean> = {
  data: true,
  bookmark: '',
  next: false,
  skip: 0,
}

const mockQueryResultMockedEcogestures: QueryResult<Ecogesture[]> = {
  data: mockedEcogesturesData,
  bookmark: '',
  next: false,
  skip: 0,
}

const mockQueryResultEmpty: QueryResult<Ecogesture[]> = {
  data: [],
  bookmark: '',
  next: false,
  skip: 0,
}

describe('Ecogesture service', () => {
  const ecogestureService = new EcogestureService(mockClient)
  describe('getAllEcogestures', () => {
    it('should return all ecogestures', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResultMockedEcogestures)
      const result = await ecogestureService.getAllEcogestures()
      expect(result).toEqual(mockedEcogesturesData)
    })
    it('should return empty array when no ecogestures stored', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResultEmpty)
      const result = await ecogestureService.getAllEcogestures()
      expect(result).toEqual([])
    })
  })

  describe('deleteAllEcogestures', () => {
    it('should return true when 3 ecogestures stored', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResultMockedEcogestures)
      const result = await ecogestureService.deleteAllEcogestures()
      expect(mockClient.destroy).toHaveBeenCalledTimes(3)
      expect(result).toBe(true)
    })
    it('should return true when no ecogestures stored', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResultEmpty)
      const result = await ecogestureService.deleteAllEcogestures()
      expect(result).toBe(true)
    })
    it('should throw exception when error happened on deletion', async () => {
      mockClient.destroy.mockRejectedValue(new Error())
      mockClient.query.mockResolvedValueOnce(mockQueryResultMockedEcogestures)
      await expect(ecogestureService.deleteAllEcogestures()).rejects.toThrow(
        new Error()
      )
    })
  })
  describe('reinitAllEcogestures', () => {
    it('should return true when 3 ecogestures stored', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResultMockedEcogestures)
      const result = await ecogestureService.reinitAllEcogestures()
      expect(mockClient.save).toHaveBeenCalledTimes(3)
      expect(result).toBe(true)
    })
    it('should return true when no ecogestures stored', async () => {
      mockClient.query.mockResolvedValueOnce(mockQueryResultEmpty)
      const result = await ecogestureService.reinitAllEcogestures()
      expect(result).toBe(true)
    })
    it('should throw exception when error happened on reinit', async () => {
      mockClient.save.mockRejectedValue(new Error())
      mockClient.query.mockResolvedValueOnce(mockQueryResultMockedEcogestures)
      await expect(ecogestureService.reinitAllEcogestures()).rejects.toThrow(
        new Error()
      )
    })
  })

  describe('filterByUsage', () => {
    it('should return ecogesture list including ECS ecogestures', async () => {
      const mockEcogestureList: Ecogesture[] = ecogesturesECSData
      const mockProfileEcogesture1: ProfileEcogesture = {
        ...mockProfileEcogesture,
        hotWater: IndividualOrCollective.INDIVIDUAL,
      }
      const result = ecogestureService.filterByUsage(
        mockEcogestureList,
        mockProfileEcogesture1
      )
      expect(result.includes(ecogesturesECSData[0])).toBeTruthy()
    })
    it('should return ecogesture list excluding ECS ecogestures', async () => {
      const mockEcogestureList: Ecogesture[] = ecogesturesECSData
      const mockProfileEcogesture1: ProfileEcogesture = {
        ...mockProfileEcogesture,
        hotWater: IndividualOrCollective.COLLECTIVE,
      }
      const result = ecogestureService.filterByUsage(
        mockEcogestureList,
        mockProfileEcogesture1
      )
      expect(result.includes(ecogesturesECSData[1])).toBeFalsy()
    })
    it('should return ecogesture list including HEATING ecogestures', async () => {
      const mockEcogestureList: Ecogesture[] = ecogesturesHeatingData
      const mockProfileEcogesture2: ProfileEcogesture = {
        ...mockProfileEcogesture,
        heating: IndividualOrCollective.INDIVIDUAL,
        warmingFluid: WarmingType.ELECTRICITY,
      }
      const result = ecogestureService.filterByUsage(
        mockEcogestureList,
        mockProfileEcogesture2
      )
      expect(result.includes(ecogesturesHeatingData[0])).toBeTruthy()
    })
    it('should return ecogesture list excluding HEATING ecogestures', async () => {
      const mockEcogestureList: Ecogesture[] = ecogesturesHeatingData
      const mockProfileEcogesture2: ProfileEcogesture = {
        ...mockProfileEcogesture,
        heating: IndividualOrCollective.COLLECTIVE,
      }
      const result = ecogestureService.filterByUsage(
        mockEcogestureList,
        mockProfileEcogesture2
      )
      expect(result.includes(ecogesturesHeatingData[0])).toBeFalsy()
    })
  })
  describe('filterByEquipment', () => {
    it('should return ecogesture list including BOILER equipment and equipment verification to true', async () => {
      const mockProfileEcogestureBOILER: ProfileEcogesture = {
        ...mockProfileEcogesture,
        equipments: [EquipmentType.BOILER],
      }
      const result = ecogestureService.filterByEquipment(
        BoilerEcogesture,
        mockProfileEcogestureBOILER
      )
      expect(result.includes(BoilerEcogesture[0])).toBeTruthy()
    })
    it('should return ecogesture list excluding BOILER equipment and equipment verification to true', async () => {
      const mockProfileEcogestureBOILER: ProfileEcogesture = {
        ...mockProfileEcogesture,
      }
      const result = ecogestureService.filterByEquipment(
        BoilerEcogesture,
        mockProfileEcogestureBOILER
      )
      expect(result.includes(BoilerEcogesture[0])).toBeFalsy()
    })
    it('should return ecogesture list including BOILER equipment with equipment verification to false but equipment to BOILER', async () => {
      const mockProfileEcogestureBOILER: ProfileEcogesture = {
        ...mockProfileEcogesture,
        equipments: [EquipmentType.BOILER],
      }
      const result = ecogestureService.filterByEquipment(
        BoilerEcogestureFalse,
        mockProfileEcogestureBOILER
      )
      expect(result.includes(BoilerEcogestureFalse[0])).toBeTruthy()
    })
  })
  describe('getEcogestureListByProfile', () => {
    it('should return ecogesture list according to profile ecogesture, sorted and filtered', async () => {
      const mockProfileEcogestureFull: ProfileEcogesture = {
        ...mockProfileEcogesture,
        equipments: [EquipmentType.WASHING_MACHINE, EquipmentType.DISHWASHER],
      }

      mockClient.query.mockResolvedValueOnce(mockQueryResultMockedEcogestures)

      const result = await ecogestureService.getEcogestureListByProfile(
        mockProfileEcogestureFull
      )
      expect(result.length).toBe(2)
      expect(result[0]).toBe(mockedEcogesturesData[0])
    })
  })
  describe('getEcogesturesByIds', () => {
    it('Should return corresponding ecogestures', async () => {
      const mockQueryResult: QueryResult<Ecogesture[]> = {
        data: ecogesturesECSData,
        bookmark: '',
        next: false,
        skip: 0,
      }
      mockClient.query.mockResolvedValueOnce(mockQueryResult)

      const result = await ecogestureService.getEcogesturesByIds([
        'ECOGESTURE0001',
        'ECOGESTURE0002',
        'ECOGESTURE0013',
      ])
      expect(result.length).toBe(3)
      expect(result).toBe(ecogesturesECSData)
    })
  })

  describe('initEcogesture()', () => {
    const hash = hashFile(ecoData)

    beforeEach(() => {
      mockClient.query.mockClear()
      mockClient.create.mockClear()
    })

    it('should return hash when ecogestures hash is already up to date', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValueOnce(ecoData)
      const { ecogestureHash } = await ecogestureService.initEcogesture(hash)
      expect(ecogestureHash).toEqual(hash)
    })
    it('should match hash and ecogesture number when ecogestures are created', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValueOnce([])
        .mockResolvedValueOnce(ecoData)
      mockClient.create.mockResolvedValue(mockQueryResultCreated)

      const { ecogestureHash, ecogestureList } =
        await ecogestureService.initEcogesture(hash)
      expect(ecogestureHash).toEqual(hash)
      expect(ecogestureList.length).toEqual(ecogestureData.length)
    })
    it('should throw an error when ecogestures should be created and created ecogestures number does not match', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValueOnce([])
        .mockResolvedValueOnce(mockedEcogesturesData)

      await expect(
        ecogestureService.initEcogesture(hashFile(ecogestureData))
      ).rejects.toThrow(
        new Error(
          'initEcogesture: Created ecogesture type entities does not match'
        )
      )
    })
    it('should throw an error when ecogestures should be created and creation failed', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValueOnce([])
      mockClient.create.mockRejectedValue(new Error())
      await expect(
        ecogestureService.initEcogesture(hashFile(ecogestureData))
      ).rejects.toThrow(new Error())
    })
    it('should return hash and ecogesture list when ecogestures are created', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValue(ecoData)
      jest
        .spyOn(ecogestureService, 'deleteAllEcogestures')
        .mockResolvedValue(true)
      mockClient.create.mockResolvedValue(mockQueryResultCreated)

      const { ecogestureHash, ecogestureList } =
        await ecogestureService.initEcogesture('')
      expect(ecogestureHash).toEqual(hash)
      expect(ecogestureList.length).toEqual(ecogestureData.length)
    })
    it('should throw an error when ecogestures should be updated and created ecogestures number does not match', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValueOnce(ecoData)
        .mockResolvedValueOnce(mockedEcogesturesData)
      jest
        .spyOn(ecogestureService, 'deleteAllEcogestures')
        .mockResolvedValue(true)
      mockClient.create.mockResolvedValue(mockQueryResultCreated)
      await expect(ecogestureService.initEcogesture('')).rejects.toThrow(
        new Error(
          'initEcogesture: Created ecogesture type entities does not match'
        )
      )
    })
    it('should throw an error when ecogestures should be updated and ecogestures creation failed', async () => {
      jest
        .spyOn(ecogestureService, 'getAllEcogestures')
        .mockResolvedValue(ecoData)
      jest
        .spyOn(ecogestureService, 'deleteAllEcogestures')
        .mockResolvedValue(true)
      mockClient.create.mockRejectedValueOnce(new Error())
      const error = await getError(async () =>
        ecogestureService.initEcogesture('')
      )
      expect(error).toEqual(new Error())
    })
  })
})
