import * as Sentry from '@sentry/react'
import { Client } from 'cozy-client'
import logger from 'cozy-logger'
import {
  DaccEvent,
  FluidSlugType,
  FluidState,
  FluidType,
  HotWaterFluid,
  TimeStep,
  UsageEventProperties,
  UsageEventType,
  UserChallengeState,
  WarmingType,
} from 'enums'
import { toNumber, uniq } from 'lodash'
import { DateTime } from 'luxon'
import { FluidStatus, PerformanceIndicator, UsageEvent } from 'models'
import ChallengeService from 'services/challenge.service'
import ConsumptionService from 'services/consumption.service'
import EnvironmentService from 'services/environment.service'
import FluidService from 'services/fluid.service'
import ProfileService from 'services/profile.service'
import ProfileTypeEntityService from 'services/profileTypeEntity.service'
import TermsService from 'services/terms.service'
import UsageEventService from 'services/usageEvent.service'
import { getFluidType } from 'utils/utils'
import { runService } from './service'

const logStack = logger.namespace('aggregatorUsageEvents')

interface AggregatorUsageEventsProps {
  client: Client
}

interface Indicator {
  createdBy: string
  measureName: string
  startDate: string
  value: number | null
  group1?: object
  group2?: object
  group3?: object
}

// Store all id of events read during the process
const readUsageEvents: UsageEvent[] = []

// Store error events
const errorEvent: { error: unknown; doctype: string }[] = []

const sendIndicator = async (
  indicator: Indicator,
  client: Client
): Promise<boolean> => {
  try {
    const environmentService = new EnvironmentService()
    logStack(
      'info',
      environmentService.isProduction()
        ? 'Sending data to dacc'
        : 'Sending data to dacc-dev'
    )

    // /!\ In order to test locally, please replace /remote/cc.cozycloud.dacc with http://localhost:8081
    await client
      .getStackClient()
      .fetchJSON(
        'POST',
        environmentService.isProduction()
          ? '/remote/cc.cozycloud.dacc_v2'
          : '/remote/cc.cozycloud.dacc.dev_v2',
        {
          data: JSON.stringify(indicator),
          path: 'measure',
        }
      )
    return true
  } catch (error) {
    let errorMessage = 'Failed to do something exceptional'
    if (error instanceof Error) {
      errorMessage = error.message
    }
    logStack(
      'error',
      `Error while sending indicator ${indicator.measureName} to remote doctype: ${errorMessage}`
    )
    Sentry.captureException(error)
    throw error
  }
}

const reduceEvents = (
  events: UsageEvent[]
): { [key: string]: UsageEvent[] } => {
  return events.reduce(function (
    acc: {
      [key: string]: UsageEvent[]
    },
    event: UsageEvent
  ) {
    const key = `${event.eventDate.startOf('day').toString()}|${event.target}|${
      event.result
    }|${event.context}`
    if (!acc[key]) {
      acc[key] = []
    }
    acc[key].push(event)
    readUsageEvents.push(event)
    return acc
  }, {})
}

const sendAggregatedEventByDay = async (
  events: UsageEvent[],
  eventType: DaccEvent,
  client: Client,
  groupsKeys: { group1: string; group2?: string; group3?: string },
  groupsIndexes: number[],
  customValues?: (string | null)[]
) => {
  const reducedEvents = reduceEvents(events)
  for (const item of Object.keys(reducedEvents)) {
    const splittedKey = item.split('|')
    let group1 = {}
    let group2 = {}
    let group3 = {}
    if (groupsKeys.group1 && (groupsIndexes[0] !== 0 || customValues?.[0])) {
      group1 = {
        [groupsKeys.group1]: customValues?.[0]
          ? customValues[0]
          : splittedKey[groupsIndexes[0]],
      }
    }
    if (groupsKeys.group2 && (groupsIndexes[1] !== 0 || customValues?.[1])) {
      group2 = {
        [groupsKeys.group2]: customValues?.[1]
          ? customValues[1]
          : splittedKey[groupsIndexes[1]],
      }
    }
    if (groupsKeys.group3 && (groupsIndexes[2] !== 0 || customValues?.[2])) {
      group3 = {
        [groupsKeys.group3]: customValues?.[2]
          ? customValues[2]
          : splittedKey[groupsIndexes[2]],
      }
    }
    const indicator: Indicator = {
      createdBy: 'ecolyo',
      measureName: eventType,
      startDate: DateTime.fromISO(splittedKey[0]).toISODate(),
      value: reducedEvents[item].length,
      group1: group1,
      ...(Object.keys(group2).length > 0 && { group2: group2 }),
      ...(Object.keys(group3).length > 0 && { group3: group3 }),
    }
    const result = await sendIndicator(indicator, client)
    if (result) {
      readUsageEvents.push(...reducedEvents[item])
    }
  }
}

const handleQuizStars = async (events: UsageEvent[], client: Client) => {
  events.forEach(async event => {
    const indicator: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccEvent.QUIZ_STARS,
      startDate: event.startDate ? event.startDate.toISODate() : '',
      value: parseInt(event.result ? event.result : '0'),
      // eslint-disable-next-line camelcase
      group1: { challenge_id: event.context ? event.context : '' },
      // eslint-disable-next-line camelcase
      group2: { quiz_id: event.target ? event.target : '' },
    }
    const result = await sendIndicator(indicator, client)
    if (result) {
      readUsageEvents.push(event)
    }
  })
}

const calculSessionTime = async (events: UsageEvent[], client: Client) => {
  let startSessionDate: DateTime | null = null
  let isFirstConnection = false
  let navigationCount = 0
  for (const [index, event] of events.entries()) {
    if (event.type === UsageEventType.CONNECTION_EVENT) {
      if (
        startSessionDate &&
        index > 0 &&
        events[index - 1].type !== UsageEventType.CONNECTION_EVENT
      ) {
        const endDate = events[index - 1].eventDate
        const duration = endDate
          .diff(startSessionDate, ['seconds'])
          .toObject().seconds
        const sessionIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.SESSION_DURATION,
          startDate: startSessionDate.toISODate(),
          value: duration === undefined ? 0 : duration,
          // eslint-disable-next-line camelcase
          group1: { number_pages: navigationCount.toString() },
          // eslint-disable-next-line camelcase
          group2: { session_type: isFirstConnection ? 'first' : 'any' },
        }
        await sendIndicator(sessionIndicator, client)
        navigationCount = 0
      }
      startSessionDate = event.eventDate
      isFirstConnection =
        event.result && event.result === 'firstConnection' ? true : false
    } else if (event.type === UsageEventType.NAVIGATION_EVENT) {
      if (startSessionDate) {
        navigationCount += 1
      }
    } else if (index === events.length - 1) {
      if (startSessionDate) {
        const endDate = event.eventDate
        const duration = endDate
          .diff(startSessionDate, ['seconds'])
          .toObject().seconds
        const sessionIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.SESSION_DURATION,
          startDate: startSessionDate.toISODate(),
          value: duration === undefined ? 0 : duration,
          // eslint-disable-next-line camelcase
          group1: { number_pages: navigationCount.toString() },
          // eslint-disable-next-line camelcase
          group2: { session_type: isFirstConnection ? 'first' : 'any' },
        }
        await sendIndicator(sessionIndicator, client)
      }
    }
  }
}

const calculPeriodBetweenFirstConnectionAndFirstChallenge = async (
  events: UsageEvent[],
  firstConnectionEvent: UsageEvent,
  client: Client
) => {
  const challengeLaunchEvents: UsageEvent[] = events.filter(
    (event: UsageEvent) => event.type === UsageEventType.CHALLENGE_LAUNCH_EVENT
  )
  if (
    challengeLaunchEvents.length > 0 &&
    challengeLaunchEvents[0].target === 'CHALLENGE0001'
  ) {
    const periodIndicator: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccEvent.EVENT_DURATION,
      startDate: firstConnectionEvent.eventDate.toISODate(),
      value: challengeLaunchEvents[0].eventDate.diff(
        firstConnectionEvent.eventDate
      ).seconds,
      // eslint-disable-next-line camelcase
      group1: { start_event: 'first_session' },
      // eslint-disable-next-line camelcase
      group2: { end_event: 'first_challenge' },
      group3: { params: '' },
    }
    const result = await sendIndicator(periodIndicator, client)
    if (result) {
      readUsageEvents.push(challengeLaunchEvents[0])
    }
  }
}

const calculPeriodBetweenChallenge = async (
  events: UsageEvent[],
  client: Client
) => {
  const challengeLaunchEvents: UsageEvent[] = events.filter(
    (event: UsageEvent) => event.type === UsageEventType.CHALLENGE_LAUNCH_EVENT
  )
  if (challengeLaunchEvents.length > 0) {
    const allEndedChallengeEvents: UsageEvent[] =
      await UsageEventService.getEvents(client, {
        type: UsageEventType.CHALLENGE_END_EVENT,
      })
    for (const event of challengeLaunchEvents) {
      if (event.target && event.target !== 'CHALLENGE0001') {
        const challengeId: number = toNumber(
          event.target.substr(event.target.length - 4)
        )
        const prevChallengeId = `CHALLENGE${(challengeId - 1)
          .toString()
          .padStart(4, '0')}`
        const previousEndedChallengeIndex: number =
          allEndedChallengeEvents.findIndex(
            (endedEvent: UsageEvent) => endedEvent.target === prevChallengeId
          )
        if (previousEndedChallengeIndex > -1) {
          const periodChallengeIndicator: Indicator = {
            createdBy: 'ecolyo',
            measureName: DaccEvent.EVENT_DURATION,
            startDate:
              allEndedChallengeEvents[
                previousEndedChallengeIndex
              ].eventDate.toISODate(),
            value: event.eventDate.diff(
              allEndedChallengeEvents[previousEndedChallengeIndex].eventDate
            ).seconds,
            // eslint-disable-next-line camelcase
            group1: { start_event: 'first_session' },
            // eslint-disable-next-line camelcase
            group2: { end_event: 'first_challenge' },
            group3: { params: event.target + ':' + prevChallengeId },
          }
          await sendIndicator(periodChallengeIndicator, client)
        }
      }
    }
  }
}

/**
 * Format a string with all konnectors in success state
 * @param konnectorFluidTypes FluidType[]
 * @returns string
 */
const getConnectedKonnectorSlug = (
  konnectorFluidTypes: FluidType[]
): string => {
  let slug = ''
  if (konnectorFluidTypes.includes(FluidType.ELECTRICITY)) {
    slug += 'electricity'
  }
  if (konnectorFluidTypes.includes(FluidType.GAS)) {
    if (slug.length > 0) {
      slug += ':'
    }
    slug += 'gas'
  }
  if (konnectorFluidTypes.includes(FluidType.WATER)) {
    if (slug.length > 0) {
      slug += ':'
    }
    slug += 'water'
  }
  return slug
}

const calculateConnectedKonnectorPerDay = async (client: Client) => {
  const fluidService: FluidService = new FluidService(client)
  const fluidStatus: FluidStatus[] = await fluidService.getFluidStatus()
  const connectedKonnectors = fluidStatus.filter(
    fluid => fluid.status === FluidState.DONE
  )
  logStack('info', 'calculateConnectedKonnectorPerDay')
  if (connectedKonnectors.length > 0) {
    const konnectorFluidTypes: FluidType[] = []
    for (const konnector of connectedKonnectors) {
      konnectorFluidTypes.push(konnector.fluidType)
    }
    const KonnectorConnectedPerDayIndicator: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccEvent.KONNECTOR_CONNECTED_PER_DAY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: connectedKonnectors.length,
      group1: { categories: getConnectedKonnectorSlug(konnectorFluidTypes) },
    }
    await sendIndicator(KonnectorConnectedPerDayIndicator, client)
  }
}

/**
 * Build indicator group string with handling of : ECS, heating for gas and electricity.
 * @param client Client
 * @param fluidType FluidType
 * @returns Promise<string>
 */
const buildProfileWithFluidType = async (
  client: Client,
  fluidType: FluidType
): Promise<string> => {
  let formattedProfile = ''
  const profile = await new ProfileService(client).getProfile()
  // If profile is not filled, return empty string

  if (!profile?.isProfileTypeCompleted) return formattedProfile
  const date = DateTime.local().setZone('utc', {
    keepLocalTime: true,
  })
  const profileType = await new ProfileTypeEntityService(client).getProfileType(
    date
  )
  if (fluidType === FluidType.ELECTRICITY) {
    if (profile && profileType?.warmingFluid === WarmingType.ELECTRICITY) {
      formattedProfile = 'chauffage élec'
    }
    if (profile && profileType?.warmingFluid === WarmingType.WOOD) {
      formattedProfile = 'chauffage bois'
    }
    if (profile && profileType?.warmingFluid === WarmingType.FUEL) {
      formattedProfile = 'chauffage fuel'
    }
    if (profile && profileType?.hotWaterFluid === HotWaterFluid.ELECTRICITY) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'ECS'
      } else {
        formattedProfile += ':ECS'
      }
    }
    if (profile && profileType?.cookingFluid === FluidType.ELECTRICITY) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'cuisine élec'
      } else {
        formattedProfile += ':cuisine élec'
      }
    }
    if (formattedProfile.length === 0) {
      return 'électricité spécifique'
    } else {
      return formattedProfile
    }
  } else if (fluidType === FluidType.GAS) {
    if (profile && profileType?.warmingFluid === WarmingType.GAS) {
      formattedProfile = 'chauffage gaz'
    }
    if (
      profile &&
      profileType &&
      profileType.hotWaterFluid === HotWaterFluid.GAS
    ) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'ECS'
      } else {
        formattedProfile += ':ECS'
      }
    }
    if (profile && profileType && profileType.cookingFluid === FluidType.GAS) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'cuisine gaz'
      } else {
        formattedProfile += ':cuisine gaz'
      }
    }
    if (formattedProfile.length === 0) {
      return 'autre'
    } else {
      return formattedProfile
    }
  }
  return formattedProfile
}

const getConsumptionValue = async (
  client: Client,
  fluidType: FluidType[]
): Promise<PerformanceIndicator[]> => {
  const consumptionService = new ConsumptionService(client)
  const analysisDate = DateTime.local().setZone('utc', { keepLocalTime: true })
  const periods = {
    timePeriod: {
      startDate: analysisDate.minus({ month: 1 }).startOf('month'),
      endDate: analysisDate.minus({ month: 1 }).endOf('month'),
    },
    comparisonTimePeriod: {
      startDate: analysisDate.minus({ month: 2 }).startOf('month'),
      endDate: analysisDate.minus({ month: 2 }).endOf('month'),
    },
  }
  const fetchedPerformanceIndicators =
    await consumptionService.getPerformanceIndicators(
      periods.timePeriod,
      TimeStep.MONTH,
      fluidType,
      periods.comparisonTimePeriod
    )
  return fetchedPerformanceIndicators
}

/**
 * Send an indicator on the consumption variation in % for each fluid type.
 * @description [{ slug }, { seniority (in month) }, { profile (ECS, chauffage, etc...) }],
 */
const calculateConsumptionVariation = async (client: Client) => {
  logStack('info', `calculateConsumptionVariation`)
  const consumptionData = await getConsumptionValue(client, [
    FluidType.ELECTRICITY,
    FluidType.GAS,
    FluidType.WATER,
  ])
  for (const fluidType in [
    FluidType.ELECTRICITY,
    FluidType.GAS,
    FluidType.WATER,
  ]) {
    if (fluidType < FluidType.MULTIFLUID.toFixed()) {
      // Seniority process
      const [firstConnectionEvent] = await UsageEventService.getEvents(client, {
        type: UsageEventType.CONNECTION_EVENT,
        result: 'firstConnection',
      })

      if (firstConnectionEvent) {
        const seniority = 0
        // Seniority is set to 0, otherwise the indicator is too specific
        //  DateTime.local()
        //       .setZone('utc', {
        //         keepLocalTime: true,
        //       })
        //       .diff(firstConnectionEvent.eventDate, 'month')
        //       .get('month')

        const consumptionVariationIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.CONSUMPTION_VARIATION_MONTHLY,
          startDate: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .startOf('day')
            .toISODate(),
          value: consumptionData[fluidType]?.percentageVariation
            ? consumptionData[fluidType].percentageVariation
            : 0, // in percent
          // eslint-disable-next-line camelcase
          group1: { fluid_type: FluidType[fluidType].toLowerCase() },
          // eslint-disable-next-line camelcase
          group2: { seniority: Math.round(seniority).toString() },
          group3: {
            // eslint-disable-next-line camelcase
            fluid_usage: await buildProfileWithFluidType(
              client,
              getFluidType(FluidType[fluidType])
            ),
          },
        }

        // if user wasn't connected during current month, dont send indicator
        const events: UsageEvent[] = await UsageEventService.getEvents(client, {
          eventDate: {
            $lt: DateTime.local()
              .setZone('utc', {
                keepLocalTime: true,
              })
              .endOf('month')
              .minus({ month: 1 })
              .toString(),
            $gt: DateTime.local()
              .setZone('utc', {
                keepLocalTime: true,
              })
              .startOf('month')
              .minus({ month: 1 })
              .toString(),
          },
        })

        if (events.length > 0 && consumptionVariationIndicator.value !== 0) {
          logStack(
            'info',
            `Send variation indicator : ${JSON.stringify(
              consumptionVariationIndicator
            )}`
          )
          await sendIndicator(consumptionVariationIndicator, client)
        }
      }
    }
  }
}

const sendConnectionCount = async (client: Client) => {
  logStack('info', `sendConnectionCount`)
  // Get daily connexion
  const events: UsageEvent[] = await UsageEventService.getEvents(client, {
    type: UsageEventType.CONNECTION_EVENT,
    eventDate: {
      $lt: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .endOf('month')
        .minus({ month: 1 })
        .toString(),
      $gt: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('month')
        .minus({ month: 1 })
        .toString(),
    },
  })

  let uniqueDates = 0
  // If there is at least one connection, filter each unique connection in order to send it
  if (events.length > 0) {
    uniqueDates = events
      .map(s => s.eventDate.day)
      .filter((s, i, a) => a.indexOf(s) == i).length
  }

  const connectionMonthly: Indicator = {
    createdBy: 'ecolyo',
    measureName: DaccEvent.CONNECTION_COUNT_MONTHLY,
    startDate: DateTime.local()
      .setZone('utc', {
        keepLocalTime: true,
      })
      .startOf('day')
      .toISODate(),
    value: uniqueDates,
  }
  logStack(
    'info',
    `Send connectionMonthly indicator : ${JSON.stringify(connectionMonthly)}`
  )
  await sendIndicator(connectionMonthly, client)
}

const sendProfileCount = async (client: Client) => {
  logStack('info', `sendProfileCount`)
  // Get profile setEvents
  const events: UsageEvent[] = await UsageEventService.getEvents(client, {
    type: UsageEventType.PROFILE_SET_EVENT,
  })

  // If there is at least one connection, filter each unique connection in order to send it
  if (events.length > 0) {
    const profileSet: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccEvent.PROFILE_COUNT_MONTHLY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: events.length,
    }
    await sendIndicator(profileSet, client)
  }
}

const sendEmailSubscription = async (client: Client) => {
  logStack('info', `sendEmailSubscription`)
  const profile = await new ProfileService(client).getProfile()
  if (profile?.sendAnalysisNotification) {
    const cameBackFromEmail: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccEvent.SUMMARY_SUBSCRIPTION_MONTHLY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: profile.sendAnalysisNotification ? 1 : 0,
      // eslint-disable-next-line camelcase
      group1: { event_type: 'subscribed' },
    }
    await sendIndicator(cameBackFromEmail, client)
  }
}

/**
 * Send indicator to dacc, if user has half-hour data.
 * @param client CozyClient
 */
const sendHalfHourConsumption = async (client: Client) => {
  logStack('info', `sendHalfHourConsumption`)
  const consumptionService = new ConsumptionService(client)

  const data = await consumptionService.getLastHourData(
    client,
    DateTime.local().minus({ month: 1 }).startOf('month').month
  )

  const halfHourConsumption: Indicator = {
    createdBy: 'ecolyo',
    measureName: DaccEvent.FLUID_DATA_GRANULARITY,
    // eslint-disable-next-line camelcase
    group1: { fluid_type: 'electricity' },
    // eslint-disable-next-line camelcase
    group2: { granularity: 'half_hour' },
    startDate: DateTime.local()
      .setZone('utc', {
        keepLocalTime: true,
      })
      .startOf('day')
      .toISODate(),
    value: data?.length > 0 ? 1 : 0,
  }
  await sendIndicator(halfHourConsumption, client)
}

/**
 * Send indicator to dacc, each month send if user has succeeded to configure a connector with the number of try.
 * @param client CozyClient
 */
const sendKonnectorEvents = async (client: Client) => {
  logStack('info', `sendKonnectorEvents`)
  const slugs = Object.values(FluidSlugType)
  const today = DateTime.local().setZone('utc', {
    keepLocalTime: true,
  })
  slugs.forEach(async slug => {
    let successEvents: UsageEvent[] = await UsageEventService.getEvents(
      client,
      {
        type: UsageEventType.KONNECTOR_CONNECT_EVENT,
        target: slug,
        result: 'success',
        eventDate: {
          $lte: today.endOf('month').minus({ month: 1 }).toString(),
          $gte: today.startOf('month').minus({ month: 1 }).toString(),
        },
      },
      true
    )

    // If there is no success in month, send nothing
    if (successEvents && successEvents.length > 0) {
      // Get all success events
      successEvents = await UsageEventService.getEvents(
        client,
        {
          type: UsageEventType.KONNECTOR_CONNECT_EVENT,
          target: slug,
          result: 'success',
        },
        true
      )
      // Remove success from other month, they should have been already processed
      // successEvents.length = successEventsOfCurrentMonth

      for (let index = 0; index < successEvents.length; index++) {
        const successEvent = successEvents[index]
        let query = null

        // If there is a previous value take it as reference for the query
        // Else get all previous because it's the first one
        if (index + 1 < successEvents.length) {
          query = {
            type: UsageEventType.KONNECTOR_CONNECT_EVENT,
            target: slug,
            result: 'error',
            eventDate: {
              $lte: successEvent.eventDate,
              $gte: successEvents[index + 1].eventDate,
            },
          }
        } else {
          query = {
            type: UsageEventType.KONNECTOR_CONNECT_EVENT,
            target: slug,
            result: 'error',
            eventDate: {
              $lte: successEvent.eventDate,
            },
          }
        }

        const allConnectionEvents: UsageEvent[] =
          await UsageEventService.getEvents(client, query, true)

        const konnectorSuccess: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.PARTNER_SUCCESS_MONTHLY,
          // eslint-disable-next-line camelcase
          group1: { fluid_type: slug },
          startDate: DateTime.local()
            .setZone('utc', {
              keepLocalTime: true,
            })
            .startOf('day')
            .toISODate(),
          value: allConnectionEvents.length + 1, // +1 in order to count the success
        }
        // Send indicator if it's in current month
        if (successEvent.eventDate.month === today.minus({ month: 1 }).month) {
          await sendIndicator(konnectorSuccess, client)
        }
      }
    }
  })
}

/**
 * Send the total number of partner connection attempts and the number of success
 * @param client CozyClient
 */
const sendKonnectorAttemptsMonthly = async (client: Client) => {
  logStack('info', `sendkonnectorAttemptsMonthly`)
  const slugs = Object.values(FluidSlugType)
  const today = DateTime.local().setZone('utc', {
    keepLocalTime: true,
  })
  // Count the number of connection and refresh events
  slugs.forEach(async slug => {
    const konnectorEvents: UsageEvent[] = await UsageEventService.getEvents(
      client,
      {
        type: UsageEventType.KONNECTOR_ATTEMPT_EVENT,
        target: slug,
        eventDate: {
          $lte: today.endOf('month').minus({ month: 1 }).toString(),
          $gte: today.startOf('month').minus({ month: 1 }).toString(),
        },
      },
      true
    )
    logStack('info', ` : ${JSON.stringify(konnectorEvents)}`)

    // Check if there is a success (will be false or true since the event is triggered only for the first connexion)
    const success: boolean =
      konnectorEvents.filter(event => event.result == 'success').length > 0

    const konnectorAttempts: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccEvent.UNINITIALIZED_KONNECTOR_ATTEMPTS_MONTHLY,
      // eslint-disable-next-line camelcase
      group1: { slug: slug },
      group2: { success: success },
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: konnectorEvents.length,
    }
    // Send indicator if there is connection events
    if (konnectorEvents.length > 0) {
      await sendIndicator(konnectorAttempts, client)
    }
  })
}

const aggregateEvents = async (
  events: UsageEvent[],
  eventType: UsageEventType,
  firstConnectionEvent: UsageEvent,
  client: Client
) => {
  let reducedEvents = null // Used to store reducedEvents during calculation
  switch (eventType) {
    case UsageEventType.CONNECTION_EVENT:
      await sendAggregatedEventByDay(
        events,
        DaccEvent.CONNECTION_COUNT_DAILY,
        client,
        { group1: 'device' },
        [UsageEventProperties.CONTEXT]
      )
      break
    case UsageEventType.KONNECTOR_CONNECT_EVENT:
      await sendAggregatedEventByDay(
        events,
        DaccEvent.KONNECTOR_EVENT_DAILY,
        client,
        { group1: 'slug', group2: 'event_type', group3: 'status' },
        [UsageEventProperties.TARGET, 0, UsageEventProperties.RESULT],
        [null, 'connexion', null]
      )
      break
    case UsageEventType.KONNECTOR_REFRESH_EVENT:
      await sendAggregatedEventByDay(
        events,
        DaccEvent.KONNECTOR_EVENT_DAILY,
        client,
        { group1: 'slug', group2: 'event_type', group3: 'status' },
        [UsageEventProperties.TARGET, 0, UsageEventProperties.RESULT],
        [null, 'refresh', null]
      )
      break
    case UsageEventType.NAVIGATION_EVENT:
      await sendAggregatedEventByDay(
        events,
        DaccEvent.NAVIGATION_COUNT_DAILY,
        client,
        { group1: 'page' },
        [UsageEventProperties.TARGET]
      )
      break
    case UsageEventType.PROFILE_SET_EVENT:
      await sendAggregatedEventByDay(
        events,
        DaccEvent.NAVIGATION_ACTION_DAILY,
        client,
        { group1: 'page', group2: 'feature', group3: 'params' },
        [0, 0, 0],
        ['profile', 'profile_completed', '{}']
      )
      break
    case UsageEventType.CONSUMPTION_COMPARE_EVENT:
      reducedEvents = reduceEvents(events)
      for (const item of Object.keys(reducedEvents)) {
        const splitedKey = item.split('|')
        const indicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.NAVIGATION_ACTION_DAILY,
          startDate: DateTime.fromISO(splitedKey[0]).toISODate(),
          value: reducedEvents[item].length,
          group1: {
            page: 'consumption',
          },
          group2: {
            feature: 'compare',
          },
          group3: {
            params:
              splitedKey[UsageEventProperties.CONTEXT] +
              ':' +
              splitedKey[UsageEventProperties.TARGET],
          },
        }
        const result = await sendIndicator(indicator, client)
        if (result) {
          readUsageEvents.push(...reducedEvents[item])
        }
      }
      break
    case UsageEventType.CONSUMPTION_CHANGE_TIMESTEP_EVENT:
      reducedEvents = reduceEvents(events)
      for (const item of Object.keys(reducedEvents)) {
        const splitedKey = item.split('|')
        const indicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.NAVIGATION_ACTION_DAILY,
          startDate: DateTime.fromISO(splitedKey[0]).toISODate(),
          value: reducedEvents[item].length,
          group1: {
            page: 'consumption',
          },
          group2: {
            feature: 'changestep',
          },
          group3: {
            params:
              splitedKey[UsageEventProperties.CONTEXT] +
              ':' +
              splitedKey[UsageEventProperties.TARGET],
          },
        }
        const result = await sendIndicator(indicator, client)
        if (result) {
          readUsageEvents.push(...reducedEvents[item])
        }
      }
      break
    case UsageEventType.ACTION_CHANGE_EVENT:
      reducedEvents = reduceEvents(events)
      for (const item of Object.keys(reducedEvents)) {
        const splitedKey = item.split('|')
        const indicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.NAVIGATION_ACTION_DAILY,
          startDate: DateTime.fromISO(splitedKey[0]).toISODate(),
          value: reducedEvents[item].length,
          group1: {
            page: 'action',
          },
          group2: {
            feature: 'change_ecogesture',
          },
          group3: {
            params:
              splitedKey[UsageEventProperties.TARGET] +
              ':' +
              splitedKey[UsageEventProperties.CONTEXT],
          },
        }
        const result = await sendIndicator(indicator, client)
        if (result) {
          readUsageEvents.push(...reducedEvents[item])
        }
      }
      break
    case UsageEventType.CHALLENGE_END_EVENT:
      for (const event of events) {
        const duelLaunchIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.EVENT_DURATION,
          startDate: event.eventDate.toISODate(),
          value: event.startDate
            ? event.eventDate.diff(event.startDate, ['seconds']).seconds
            : 0,
          // eslint-disable-next-line camelcase
          group1: { start_event: 'start_challenge' },
          // eslint-disable-next-line camelcase
          group2: { end_event: 'end_challenge' },
          group3: { params: event.target ? event.target : '' },
        }
        const result = await sendIndicator(duelLaunchIndicator, client)

        if (result) {
          readUsageEvents.push(event)
        }
      }
      break
    case UsageEventType.QUIZ_END_EVENT:
      handleQuizStars(events, client)
      break
    case UsageEventType.EXPLORATION_END_EVENT:
      for (const event of events) {
        const duelLaunchIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.EVENT_DURATION,
          startDate: event.eventDate.toISODate(),
          value: event.startDate
            ? event.eventDate.diff(event.startDate, ['seconds']).seconds
            : 0,
          // eslint-disable-next-line camelcase
          group1: { start_event: 'start_exploration' },
          // eslint-disable-next-line camelcase
          group2: { end_event: 'end_exploration' },
          group3: { params: event.context + ':' + event.target },
        }
        const result = await sendIndicator(duelLaunchIndicator, client)

        if (result) {
          readUsageEvents.push(event)
        }
      }
      break
    case UsageEventType.ACTION_END_EVENT:
      for (const event of events) {
        const duelLaunchIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.EVENT_DURATION,
          startDate: event.eventDate.toISODate(),
          value: event.startDate
            ? event.eventDate.diff(event.startDate, ['seconds']).seconds
            : 0,
          // eslint-disable-next-line camelcase
          group1: { start_event: 'start_action' },
          // eslint-disable-next-line camelcase
          group2: { end_event: 'end_action' },
          group3: { params: event.target + ':' + event.context },
        }
        const result = await sendIndicator(duelLaunchIndicator, client)

        if (result) {
          readUsageEvents.push(event)
        }
      }
      break
    case UsageEventType.DUEL_END_EVENT:
      for (const event of events) {
        const duelLaunchIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.EVENT_DURATION,
          startDate: event.eventDate.toISODate(),
          value: event.startDate
            ? event.eventDate.diff(event.startDate, ['seconds']).seconds
            : 0,
          // eslint-disable-next-line camelcase
          group1: { start_event: 'start_duel' },
          // eslint-disable-next-line camelcase
          group2: { end_event: 'end_duel' },
          group3: {
            params: event.target + ':' + event.context + ':' + event.result,
          },
        }
        const result = await sendIndicator(duelLaunchIndicator, client)

        if (result) {
          readUsageEvents.push(event)
        }
      }
      break
    case UsageEventType.DUEL_LAUNCH_EVENT:
      // Get challenge ID
      const challenges = await new ChallengeService(
        client
      ).getAllUserChallengeEntities()
      const filteredCurrentChallenge = challenges.filter(
        challenge =>
          challenge.state === UserChallengeState.ONGOING ||
          challenge.state === UserChallengeState.DUEL
      )
      const currentChallenge = filteredCurrentChallenge[0]
        ? filteredCurrentChallenge[0]
        : null

      let starNumbers = 0
      if (currentChallenge?.progress) {
        starNumbers =
          currentChallenge.progress.actionProgress +
          currentChallenge.progress.explorationProgress +
          currentChallenge.progress.quizProgress
      }

      for (const event of events) {
        const duelLaunchIndicator: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.EVENT_DURATION,
          startDate: event.eventDate.toISODate(),
          value: event.startDate
            ? event.eventDate.diff(event.startDate, ['seconds']).seconds
            : 0,
          // eslint-disable-next-line camelcase
          group1: { start_event: 'start_duel' },
          // eslint-disable-next-line camelcase
          group2: { end_event: 'start_duel' },
          group3: {
            params: event.target + ':' + event.context + ':' + starNumbers,
          },
        }
        const result = await sendIndicator(duelLaunchIndicator, client)

        if (result) {
          readUsageEvents.push(event)
        }
      }
      break
    case UsageEventType.CHALLENGE_LAUNCH_EVENT:
      await calculPeriodBetweenFirstConnectionAndFirstChallenge(
        events,
        firstConnectionEvent,
        client
      )
      await calculPeriodBetweenChallenge(events, client)
      break
    case UsageEventType.REPORT_FROM_EMAIL:
      for (const event of events) {
        const cameBackFromEmail: Indicator = {
          createdBy: 'ecolyo',
          measureName: DaccEvent.SUMMARY_SUBSCRIPTION_MONTHLY,
          startDate: event.eventDate.toISODate(),
          value: 1,
          // eslint-disable-next-line camelcase
          group1: { event_type: 'email' },
        }
        const result = await sendIndicator(cameBackFromEmail, client)

        if (result) {
          readUsageEvents.push(event)
        }
      }
      break
    default:
      break
  }
}

const AggregatorUsageEvents = async ({
  client,
}: AggregatorUsageEventsProps) => {
  logStack('info', 'Launch service')
  logStack('info', 'Dacc consent validation...')
  const termService = new TermsService(client)
  if (!(await termService.getLastTerm())) {
    logStack('info', 'Exit: no consent available')
    return
  }

  const events: UsageEvent[] = await UsageEventService.getEvents(client, {
    aggregated: false,
    eventDate: {
      $lt: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day'),
    },
  })
  const [firstConnectionEvent] = await UsageEventService.getEvents(client, {
    type: UsageEventType.CONNECTION_EVENT,
    result: 'firstConnection',
  })
  if (events.length > 0) {
    logStack('info', `Fetching Sessions`)
    calculSessionTime(events, client)
    for (const eventType of Object.values(UsageEventType)) {
      logStack('info', `Fetching ${eventType}`)
      const filteredEvents: UsageEvent[] = events.filter(
        event => event.type === eventType
      )
      if (filteredEvents.length > 0) {
        await aggregateEvents(
          filteredEvents,
          eventType,
          firstConnectionEvent,
          client
        )
      }
    }
  }
  // Daily events
  calculateConnectedKonnectorPerDay(client)
  // Monthly events
  const profile = await new ProfileService(client).getProfile()
  if (
    profile &&
    DateTime.local()
      .setZone('utc', {
        keepLocalTime: true,
      })
      .startOf('day').day === profile.monthlyAnalysisDate.day
  ) {
    await sendKonnectorAttemptsMonthly(client)
    await calculateConsumptionVariation(client)
    await sendEmailSubscription(client)
    await sendHalfHourConsumption(client)
    await sendKonnectorEvents(client)
    await sendConnectionCount(client)
    await sendProfileCount(client)
  }

  const uniqueReadUsageEvents: UsageEvent[] = uniq(readUsageEvents)
  logStack(
    'info',
    `Tag aggregated usage: total of ${uniqueReadUsageEvents.length} events`
  )
  await UsageEventService.updateUsageEventsAggregated(
    client,
    uniqueReadUsageEvents
  )

  if (errorEvent.length > 0) {
    // Build error
    let error = `Process end with ${errorEvent.length} errors : `
    errorEvent.forEach(el => {
      error += `${el.doctype}, `
    })
    logStack('error', error)
    Sentry.captureException(error)
    throw error
  }

  logStack('info', 'End of service')
}

runService(AggregatorUsageEvents)
